#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import io
import os
from os.path import join as pjoin
from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist
from setuptools.command.develop import develop
from setuptools.command.bdist_egg import bdist_egg
import sys

try:
    from wheel.bdist_wheel import bdist_wheel
except ImportError:
    bdist_wheel = None


# ---------------------------------------------------------------------------
# Top Level Variables
# ---------------------------------------------------------------------------

here = os.path.abspath(os.path.dirname(__file__))
is_repo = os.path.exists(pjoin(here, '.git'))
name = 'jupyterlab_server'

version_ns = {}
with io.open(pjoin(here, name, '_version.py'), encoding='utf8') as f:
    exec(f.read(), {}, version_ns)
__version__ = version_ns['__version__']


# ---------------------------------------------------------------------------
# Public Functions
# ---------------------------------------------------------------------------

def create_cmdclass(data_dirs=None):
    """Create a command class with the given optional wrappers.

    Parameters
    ----------
    wrappers: list(str), optional
        The cmdclass names to run before running other commands
    data_dirs: list(str), optional.
        The directories containing static data.
    """
    egg = bdist_egg if 'bdist_egg' in sys.argv else bdist_egg_disabled
    cmdclass = dict(
        build_py=build_py,
        sdist=sdist,
        bdist_egg=egg,
        develop=develop
    )
    if bdist_wheel:
        cmdclass['bdist_wheel'] = bdist_wheel
    return cmdclass


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg
    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    """

    def run(self):
        sys.exit('Aborting implicit building of eggs. Use `pip install .` ' +
                 ' to install from source.')



import errno
import json
import os
import sys
import requests
from contextlib import contextmanager

import tornado


here = os.path.dirname(__file__)

with open(
    os.path.join(here, 'app-settings', 'overrides.json'),
    encoding='utf-8'
) as fpt:
    big_unicode_string = json.load(fpt)["@jupyterlab/unicode-extension:plugin"]["comment"]


def maybe_patch_ioloop():
    """ a windows 3.8+ patch for the asyncio loop
    """
    if sys.platform.startswith("win") and tornado.version_info < (6, 1):
        if sys.version_info >= (3, 8):
            import asyncio
            try:
                from asyncio import (
                    WindowsProactorEventLoopPolicy,
                    WindowsSelectorEventLoopPolicy,
                )
            except ImportError:
                pass
                # not affected
            else:
                if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
                    # WindowsProactorEventLoopPolicy is not compatible with tornado 6
                    # fallback to the pre-3.8 default of Selector
                    asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def expected_http_error(error, expected_code, expected_message=None):
    """Check that the error matches the expected output error."""
    e = error.value
    if isinstance(e, tornado.web.HTTPError):
        if expected_code != e.status_code:
            return False
        if expected_message is not None and expected_message != str(e):
            return False
        return True
    elif any([
        isinstance(e, tornado.httpclient.HTTPClientError),
        isinstance(e, tornado.httpclient.HTTPError)
    ]):
        if expected_code != e.code:
            return False
        if expected_message:
            message = json.loads(e.response.body.decode())['message']
            if expected_message != message:
                return False
        return True

@contextmanager
def assert_http_error(status, msg=None):
    try:
        yield
    except requests.HTTPError as e:
        real_status = e.response.status_code
        assert real_status == status, \
                    "Expected status %d, got %d" % (status, real_status)
        if msg:
            assert msg in str(e), e
    else:
        assert False, "Expected HTTP error status"



"""Test the translations service API."""

import json
import os
import pytest
import shutil
import subprocess
import sys

from .utils import expected_http_error
from ..translation_utils import (_get_installed_language_pack_locales,
                                 _get_installed_package_locales,
                                 get_display_name,
                                 get_installed_packages_locale,
                                 get_language_pack, get_language_packs,
                                 is_valid_locale, merge_locale_data,
                                 run_process_and_parse, translator)

from .utils import maybe_patch_ioloop

maybe_patch_ioloop()

# Constants
HERE = os.path.abspath(os.path.dirname(__file__))

if not os.path.exists(os.path.join(HERE, 'translations')):
    pytest.skip("skipping translation tests", allow_module_level=True)


def setup_module(module):
    """ setup any state specific to the execution of this module."""
    for pkg in ['jupyterlab-some-package', 'jupyterlab-language-pack-es_CO']:
        src = os.path.join(HERE, 'translations', pkg)
        subprocess.Popen([sys.executable, '-m', 'pip', 'install', src]).communicate()


def teardown_module(module):
    """ teardown any state that was previously setup."""
    for pkg in ['jupyterlab-some-package', 'jupyterlab-language-pack-es_CO']:
        subprocess.Popen([sys.executable, '-m', 'pip', 'uninstall', pkg, '-y']).communicate()


@pytest.fixture(autouse=True)
def before_after_test(schemas_dir, user_settings_dir, labserverapp):

    # Code that will run before any test.

    # Copy the schema files.
    src = os.path.join(HERE, 'schemas', '@jupyterlab')
    dst = os.path.join(str(schemas_dir), '@jupyterlab')
    if os.path.exists(dst):
        shutil.rmtree(dst)

    shutil.copytree(src, dst)

    # Copy the overrides file.
    src = os.path.join(HERE, 'app-settings', 'overrides.json')
    dst = os.path.join(str(user_settings_dir), 'overrides.json')

    if os.path.exists(dst):
        os.remove(dst)

    shutil.copyfile(src, dst)

    # A test function will be run at this point.

    yield

    # Code that will run after your test.
    # N/A


async def test_get(jp_fetch):
    r = await jp_fetch("lab", "api", "translations", "")
    data = json.loads(r.body.decode())["data"]
    assert "en" in data

async def test_get_locale(jp_fetch):
    locale = "es_CO"
    r = await jp_fetch("lab", "api", "translations", locale)
    data = json.loads(r.body.decode())["data"]
    assert "jupyterlab" in data
    assert data["jupyterlab"][""]["language"] == locale

    assert "jupyterlab_some_package" in data
    assert data["jupyterlab_some_package"][""]["version"] == "0.1.0"
    assert data["jupyterlab_some_package"][""]["language"] == locale

async def test_get_locale_bad(jp_fetch):
    r = await jp_fetch("lab", "api", "translations", "foo_BAR")
    data = json.loads(r.body.decode())["data"]
    assert data == {}

async def test_get_locale_not_installed(jp_fetch):
    r = await jp_fetch("lab", "api", "translations", "es_AR")
    result = json.loads(r.body.decode())
    assert "not installed" in result["message"]
    assert result["data"] == {}

async def test_get_locale_not_valid(jp_fetch):
    r = await jp_fetch("lab", "api", "translations", "foo_BAR")
    result = json.loads(r.body.decode())
    assert "not valid" in result["message"]
    assert result["data"] == {}


# --- Backend locale
# ------------------------------------------------------------------------
async def test_backend_locale(jp_fetch):
    locale = "es_CO"
    r = await jp_fetch("lab", "api", "translations", locale)
    trans = translator.load("jupyterlab")
    result = trans.__("MORE ABOUT PROJECT JUPYTER")
    assert result == "Más sobre el proyecto jupyter"


async def test_backend_locale_extension(jp_fetch):
    locale = "es_CO"
    r = await jp_fetch("lab", "api", "translations", locale)
    trans = translator.load("jupyterlab_some_package")
    result = trans.__("BOOM")
    assert result == "Foo bar 2"


# --- Utils testing
# ------------------------------------------------------------------------
def test_get_installed_language_pack_locales_fails():
    # This should not be able to find entry points, since it needs to be
    # ran in a subprocess
    data, message = _get_installed_language_pack_locales()
    assert "es_CO" not in data
    assert message == ""

def test_get_installed_language_pack_locales_passes():
    utils_file = os.path.join(os.path.dirname(HERE), "translation_utils.py")
    cmd = [sys.executable, utils_file, "_get_installed_language_pack_locales"]
    data, message = run_process_and_parse(cmd)
    assert "es_CO" in data
    assert message == ""

def test_get_installed_package_locales_fails():
    # This should not be able to find entry points, since it needs to be
    # ran in a subprocess
    data, message = _get_installed_package_locales()
    assert "jupyterlab_some_package" not in data
    assert message == ""

def test_get_installed_package_locales():
    utils_file = os.path.join(os.path.dirname(HERE), "translation_utils.py")
    cmd = [sys.executable, utils_file, "_get_installed_package_locales"]
    data, message = run_process_and_parse(cmd)
    assert "jupyterlab_some_package" in data
    assert os.path.isdir(data["jupyterlab_some_package"])
    assert message == ""

def test_get_installed_packages_locale():
    data, message = get_installed_packages_locale("es_CO")
    assert "jupyterlab_some_package" in data
    assert "" in data["jupyterlab_some_package"]
    assert message == ""

def test_get_language_packs():
    data, message = get_language_packs("en")
    assert "en" in data
    assert "es_CO" in data
    assert message == ""

def test_get_language_pack():
    data, message = get_language_pack("es_CO")
    assert "jupyterlab" in data
    assert "jupyterlab_some_package" in data
    assert "" in data["jupyterlab"]
    assert "" in data["jupyterlab_some_package"]
    assert message == ""


# --- Utils
# ------------------------------------------------------------------------
def test_merge_locale_data():
    some_package_data_1 = {
        "": {
            "domain": "some_package",
            "version": "1.0.0"
        },
        "FOO": ["BAR"],
    }
    some_package_data_2 = {
        "": {
            "domain": "some_package",
            "version": "1.1.0"
        },
        "SPAM": ["BAR"],
    }
    some_package_data_3 = {
        "": {
            "domain": "some_different_package",
            "version": "1.4.0"
        },
        "SPAM": ["BAR"],
    }
    # Package data 2 has a newer version so it should update the package data 1
    result = merge_locale_data(some_package_data_1, some_package_data_2)
    assert "SPAM" in result
    assert "FOO" in result

    # Package data 2 has a older version so it should not update the package data 2
    result = merge_locale_data(some_package_data_2, some_package_data_1)
    assert "SPAM" in result
    assert "FOO" not in result

    # Package data 3 is a different package (domain) so it should not update package data 2
    result = merge_locale_data(some_package_data_2, some_package_data_3)
    assert result == some_package_data_2


def test_is_valid_locale_valid():
    assert is_valid_locale("en")
    assert is_valid_locale("es")
    assert is_valid_locale("es_CO")


def test_is_valid_locale_invalid():
    assert not is_valid_locale("foo_SPAM")
    assert not is_valid_locale("bar")


def test_get_display_name_valid():
    assert get_display_name("en", "en") == "English"
    assert get_display_name("en", "es") == "Inglés"
    assert get_display_name("en", "es_CO") == "Inglés"
    assert get_display_name("en", "fr") == "Anglais"
    assert get_display_name("es", "en") == "Spanish"
    assert get_display_name("fr", "en") == "French"


def test_get_display_name_invalid():
    assert get_display_name("en", "foo") == "English"
    assert get_display_name("foo", "en") == "English"
    assert get_display_name("foo", "bar") == "English"



"""Test the Settings service API.
"""

import pytest
import json
import json5
import tornado

from strict_rfc3339 import rfc3339_to_timestamp

from .utils import expected_http_error
from .utils import maybe_patch_ioloop, big_unicode_string


async def test_get(jp_fetch, labserverapp):
    id = '@jupyterlab/apputils-extension:themes'
    r = await jp_fetch('lab', 'api', 'settings', id)
    assert r.code == 200
    res = r.body.decode()
    data = json.loads(res)
    assert data['id'] == id
    schema = data['schema']
    # Check that overrides.json file is respected.
    assert schema['properties']['theme']['default'] == 'JupyterLab Dark'
    assert 'raw' in res


async def test_get_federated(jp_fetch, labserverapp):
    id = '@jupyterlab/apputils-extension-federated:themes'
    r = await jp_fetch('lab', 'api', 'settings', id)
    assert r.code == 200
    res = r.body.decode()
    assert 'raw' in res


async def test_get_bad(jp_fetch, labserverapp):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch('foo')
    assert expected_http_error(e, 404)

async def test_listing(jp_fetch, labserverapp):
    ids = [
        '@jupyterlab/apputils-extension:themes',
        '@jupyterlab/apputils-extension-federated:themes',
        '@jupyterlab/codemirror-extension:commands',
        '@jupyterlab/codemirror-extension-federated:commands',
        '@jupyterlab/shortcuts-extension:plugin',
        '@jupyterlab/translation-extension:plugin',
        '@jupyterlab/unicode-extension:plugin'
    ]
    versions = ['N/A', 'N/A', 'test-version']
    r = await jp_fetch('lab', 'api', 'settings')
    assert r.code == 200
    res = r.body.decode()
    response = json.loads(res)
    response_ids = [item['id'] for item in response['settings']]
    response_schemas = [item['schema'] for item in response['settings']]
    response_versions = [item['version'] for item in response['settings']]
    assert set(response_ids) == set(ids)
    assert all(response_schemas)
    assert set(response_versions) == set(versions)
    last_modifieds = [item['last_modified'] for item in response['settings']]
    createds = [item['created'] for item in response['settings']]
    assert {None} == set(last_modifieds + createds)


async def test_patch(jp_fetch, labserverapp):
    id = '@jupyterlab/shortcuts-extension:plugin'

    r = await jp_fetch('lab', 'api', 'settings', id,
        method='PUT',
        body=json.dumps(dict(raw=json5.dumps(dict()))))
    assert r.code == 204

    r = await jp_fetch('lab', 'api', 'settings', id,
        method='GET',
        )
    data = json.loads(r.body.decode())
    first_created = rfc3339_to_timestamp(data['created'])
    first_modified = rfc3339_to_timestamp(data['last_modified'])

    r = await jp_fetch('lab', 'api', 'settings', id,
        method='PUT',
        body=json.dumps(dict(raw=json5.dumps(dict())))
        )
    assert r.code == 204

    r = await jp_fetch('lab', 'api', 'settings', id,
        method='GET',
        )
    data = json.loads(r.body.decode())
    second_created = rfc3339_to_timestamp(data['created'])
    second_modified = rfc3339_to_timestamp(data['last_modified'])

    assert first_created <= second_created
    assert first_modified < second_modified

    r = await jp_fetch('lab', 'api', 'settings', '',
        method='GET',
        )
    data = json.loads(r.body.decode())
    listing = data['settings']
    list_data = [item for item in listing if item['id'] == id][0]
    # TODO(@echarles) Check this...
#    assert list_data['created'] == data['created']
#    assert list_data['last_modified'] == data['last_modified']


async def test_patch_unicode(jp_fetch, labserverapp):
    id = '@jupyterlab/unicode-extension:plugin'
    settings = dict(comment=big_unicode_string[::-1])
    payload = dict(raw=json5.dumps(settings))

    r = await jp_fetch('lab', 'api', 'settings', id,
        method='PUT',
        body=json.dumps(payload)
        )
    assert r.code == 204

    r = await jp_fetch('lab', 'api', 'settings', id,
        method='GET',
        )
    data = json.loads(r.body.decode())
    assert data["settings"]["comment"] == big_unicode_string[::-1]

async def test_patch_wrong_id(jp_fetch, labserverapp):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch('foo',
            method='PUT',
            body=json.dumps(dict(raw=json5.dumps(dict())))
        )
    assert expected_http_error(e, 404)

async def test_patch_bad_data(jp_fetch, labserverapp):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        settings = dict(keyMap=10)
        payload = dict(raw=json5.dumps(settings))
        await jp_fetch('foo',
            method='PUT',
            body=json.dumps(payload)
        )
    assert expected_http_error(e, 404)

async def test_patch_invalid_payload_format(jp_fetch, labserverapp):
    id = '@jupyterlab/apputils-extension:themes'

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        settings = dict(keyMap=10)
        payload = dict(foo=json5.dumps(settings))
        await jp_fetch('lab', 'api', 'settings', id,
            method='PUT',
            body=json.dumps(payload)
        )
    assert expected_http_error(e, 400)

async def test_patch_invalid_json(jp_fetch, labserverapp):
    id = '@jupyterlab/apputils-extension:themes'

    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        payload_str = 'eh'
        await jp_fetch('lab', 'api', 'settings', id,
            method='PUT',
            body=json.dumps(payload_str)
        )
    assert expected_http_error(e, 400)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from setupbase import create_cmdclass, __version__
from setuptools import find_packages, setup
import sys


setup_args = dict(
    name='jupyterlab_server',
    version=__version__,
    packages=find_packages('.'),
    description='JupyterLab Server',
    long_description=open('./README.md').read(),
    long_description_content_type='text/markdown',
    author='Jupyter Development Team',
    author_email='jupyter@googlegroups.com',
    url='https://jupyter.org',
    license='BSD',
    platforms='Linux, Mac OS X, Windows',
    keywords=['Jupyter', 'JupyterLab'],
    cmdclass=create_cmdclass(),
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    include_package_data=True
)

if 'setuptools' in sys.modules:
    setup_args['python_requires'] = '>=3.6'
    setup_args['extras_require'] = {
        'test': [
            'codecov',
            'ipykernel',
            'pytest>=5.3.2',
            'pytest-cov',
            'jupyter_server[test]',
            'pytest-console-scripts',
            'strict-rfc3339',
            'wheel',
        ],
    }
    setup_args['install_requires'] = [
        'babel',
        'jinja2>=2.10',
        'json5',
        'jsonschema>=3.0.1',
        'packaging',
        'requests',
        'jupyter_server~=1.1',
    ]

if __name__ == '__main__':
    setup(**setup_args)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""
Translation handler.
"""

import json
import os
import traceback

import tornado
from tornado import gen, web

from .settings_handler import get_settings
from .translation_utils import get_language_pack, get_language_packs, is_valid_locale, translator

from .server import APIHandler, url_path_join

SCHEMA_NAME = '@jupyterlab/translation-extension:plugin'


def get_current_locale(config):
    """
    Get the current locale for given `config`.

    Parameters
    ----------
    config: LabConfig
        The config.

    Returns
    -------
    str
        The current locale string.

    Notes
    -----
    If the locale setting is not valid, it will default to "en".
    """
    settings, _warnings = get_settings(
        config.app_settings_dir,
        config.schemas_dir,
        config.user_settings_dir,
        schema_name=SCHEMA_NAME,
    )
    current_locale = settings.get("settings", {}).get("locale", "en")
    if not is_valid_locale(current_locale):
        current_locale = "en"

    return current_locale


class TranslationsHandler(APIHandler):

    def initialize(self, lab_config):
        self.lab_config = lab_config

    @gen.coroutine
    @tornado.web.authenticated
    def get(self, locale=""):
        """
        Get installed language packs.

        Parameters
        ----------
        locale: str, optional
            If no locale is provided, it will list all the installed language packs.
            Default is `""`.
        """
        data, message = {}, ""
        try:
            if locale == "":
                data, message = get_language_packs(
                    display_locale=get_current_locale(self.lab_config))
            else:
                translator.set_locale(locale)
                data, message = get_language_pack(locale)
                if data == {} and message == "":
                    if is_valid_locale(locale):
                        message = "Language pack '{}' not installed!".format(locale)
                    else:
                        message = "Language pack '{}' not valid!".format(locale)
        except Exception:
            message = traceback.format_exc()

        self.set_status(200)
        self.finish(json.dumps({"data": data, "message": message}))
